//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <vcl\inifiles.hpp>
#include <vcl\clipbrd.hpp>
#include <filectrl.hpp>
#include <time.h>
#include <math.h>
#pragma hdrstop

#include "HashCashDlgs.h"
#include "HashCashThrds.h"
#include "HashTimeThrds.h"
#include "ImportDlgs.h"
#include "ExportDlgs.h"
#include "utct.h"
#include "EditDlgs.h"
#include "LogDlgs.h"
#include "SpendDlgs2.h"
#include "msgboxdlgs.h"
#include "AboutDlgs.h"
#include "HelpDlgs.h"
#include "global.h"
//---------------------------------------------------------------------------
#pragma link "explbtn"
#pragma resource "*.dfm"
THashCashDlg *HashCashDlg;
HWND hWndHCW;
//---------------------------------------------------------------------------
__fastcall THashCashDlg::THashCashDlg(TComponent* Owner)
	: TForm(Owner)
{
   cap = Caption;
   AnsiString priority,winpos;
   // get ini values from 'main'
   TIniFile *ini = new TIniFile(AppIni);
   winpos=ini->ReadString("main","windowpos","1,10,10");
   centered = GetToken(0,winpos,",")=="1";
   Left = GetToken(1,winpos,",").ToIntDef(10);
   Top = GetToken(2,winpos,",").ToIntDef(10);
   if (centered) CenterBtn->Checked = true;
   else RememberBtn->Checked = true;
   if (!centered) Position = poDesigned;
   QtyEdit->Text  = (int)ini->ReadInteger("main","tokens",1);
   BitsEdit->Text = (int)ini->ReadInteger("main","bits",17);
   ExpireTime->Text = ini->ReadString("main","expires","28:00:00:00");
   AnonTime->Text = ini->ReadString("main","randperiod","0");
   TimeWide->ItemIndex = TimeWide->Items->IndexOf(ini->ReadString("main","timewidth","6"));
   PriorityBox->ItemIndex = PriorityBox->Items->IndexOf(ini->ReadString("main","priority","normal"));
   Xhashcash = ini->ReadString("main","xhashcash","X-Hashcash:").Trim()+" ";
   WColorBox->ItemIndex = ini->ReadInteger("main","windowtextcolor",9);
   IColorBox->ItemIndex = ini->ReadInteger("main","inputtextcolor",4);
   Resource->Text = ini->ReadString("main","resource","fluffybunny");
   delete ini;

   PageControl->ActivePage = TabMint;
   MintDisplay->Clear();
   MintPanel->Top = 9;

   Label7->Caption = "&HashCash Tokens";
   Label8->Caption = "";
   attended = true;
   clSpend = false;
   clMint = false;
   clSetup = false;
   clPriority = "";
   clBits = -1;
   clTokens = -1;
   clResource = "";
   clMinimized = 0;
   modified = false;

   int i=0; while(++i<=ParamCount())
   {
      AnsiString arg = ParamStr(i);
      if (arg=="-setup")
      {
         clSetup = true;
         break;
      }
      if (arg=="-mint")
      {
         clMint = true;
         continue;
      }
      if (arg=="-spend")
      {
         clSpend = true;
         continue;
      }
      if (arg=="-minimized" || arg=="-m")
      {
         clMinimized = true;
         continue;
      }
      if (arg=="-tokens" || arg=="-t")
      {
         if (i+1<=ParamCount()) clTokens = ParamStr(++i).ToIntDef(-1);
         if (clTokens<=0) MsgDlg(this,
            s("The command line argument for 'tokens' is not valid!\n")+
            s("Please see help for syntax."),
            s("Data Error"),MB_OK);
         continue;
      }
      if (arg=="-bits" || arg=="-b")
      {
         if (i+1<=ParamCount()) clBits = ParamStr(++i).ToIntDef(-1);
         if (clBits<=0) MsgDlg(this,
            s("The command line argument for 'bits' is not valid!\n")+
            s("please see help for syntax."),
            s("Data Error"),MB_OK);
         continue;
      }
      if (arg=="-resource" || arg=="-r")
      {
         if (i+1<=ParamCount()) clResource = ParamStr(++i);
         if (clResource.IsEmpty()) MsgDlg(this,
            s("The command line argument for 'resource' is not valid!\n")+
            s("please see help for syntax."),
            s("Data Error"),MB_OK);
         continue;
      }
      if (arg=="-priority" || arg=="-p")
      {
         AnsiString prioritystrs = ",idle,lowest,lower,normal,higher,highest,timecritical,";
         if (i+1<=ParamCount()) clPriority = ParamStr(++i).LowerCase();
         if (prioritystrs.Pos(","+clPriority+",")==0)
            MsgDlg(this,
            s("The command line argument for 'priority' is not valid!\n")+
            s("please see help for syntax."),
            s("Data Error"),MB_OK);
      }
      /*if (arg=="-validity" || arg=="-v")
      {
         if (i+1<=ParamCount()) clValidity = ParamStr(++i).LowerCase();
         if (prioritystrs.Pos(","+clPriority+",")==0)
            MsgDlg(this,
            s("The command line argument for 'validity period' is not valid!\n")+
            s("please see help for syntax."),
            s("Data Error"),MB_OK);
      }
      if (arg=="-random" || arg=="-r")
      {
         if (i+1<=ParamCount()) clValidity = ParamStr(++i).LowerCase();
         if (prioritystrs.Pos(","+clPriority+",")==0)
            MsgDlg(this,
            s("The command line argument for 'random period' is not valid!\n")+
            s("please see help for syntax."),
            s("Data Error"),MB_OK);
      }*/
   }
   if (clSetup)
   {
      WriteDesktopShortcuts();
      clSetup = false;
      Snooze(300);
   }
   if (clSpend)
   {
      // try to avoid saving conflicts if HCW open in interactive mode as well
      WaitWhileTokenFileLocked();
      if (!FileExists(AppQueLck)) CreateEmptyFile(AppQueLck);
      bool spent=SpendToken2(this,0,"");
      if (spent || hWndHCW)
      {
         if (!spent && hWndHCW)
         {
            ShowWindow(hWndHCW,SW_SHOWDEFAULT);
            SetForegroundWindow(hWndHCW);
         }
         Application->ShowMainForm = false;
         PostMessage(Handle,WM_QUIT,0,0);
         DeleteFile(AppQueLck.c_str());
         return;
      }
      DeleteFile(AppQueLck.c_str());
      clSpend = false;
   }

   // get tokes from queue
   if (FileExists(AppQueLck)) DeleteFile(AppQueLck.c_str());
   LoadTokenPool(Pool->Items);
   if (Pool->Items->Count)
   {
      //Pool->Selected[0] = true;
      //Pool->ItemIndex = 0;
   }
   Label12->Caption = s(Pool->Items->Count)+" items";
   Label7->Caption = "&HashCash Tokens";
   Label8->Caption = "";

   WColorBoxChange(NULL);
   IColorBoxChange(NULL);
   CheckDeleteExpireds();
   AnsiString logfile = ChangeFileExt(AppExeName,".log");
   //if (FileExists(logfile)) OpenBtn->Visible;
   attended = !clMint;
   if (!attended)
   {
      Timer1->Enabled = true;
      return;
   }
}
//---------------------------------------------------------------------------
__fastcall THashCashDlg::~THashCashDlg()
{
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::WaitWhileTokenFileLocked()
{
   // waits up to 2 seconds then moves on to do allow save
   unsigned tick=GetTickCount();
   while(FileExists(AppQueLck))
   {
      if (GetTickCount()>tick+2000) break;
      AppProcessMessages();
   }
 }
//---------------------------------------------------------------------------
void  __fastcall THashCashDlg::WMSpendToken(TMessage& Msg)
{
   AnsiString token;
   int i=-1; while(++i<Pool->Items->Count)
   {
      token = Pool->Items->Strings[i];
      if (TokenStatus(token)>HASHCASH_OK)
      {
         Clipboard()->AsText = Xhashcash+": "+token;
         Msg.Result = MAKELONG(i,Pool->Items->Count);
         Pool->Items->Delete(i);
         SaveTokenPool(Pool->Items);
         return;
      }
   }
   Msg.Result = MAKELONG(65535,Pool->Items->Count);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TimeTo2Digits(TEdit *value)
{
   AnsiString result = value->Text.Trim();
   while (result.Length()<2) result = "0"+result;
   return result;
}
//---------------------------------------------------------------------------
AnsiString __fastcall GetTokenOrZero(int pos, AnsiString fromstring, AnsiString divider)
{
   AnsiString result = GetToken(pos,fromstring,divider.c_str());
   if (result.SubString(1,1)=="0") result = result.SubString(2,64);
   if (result.IsEmpty())
      return "0";
   return result;
}
//---------------------------------------------------------------------------
bool __fastcall THashCashDlg::WriteDesktopShortcuts()
{
   AnsiString AppName,TargetPath,LinkPath,IconPath,Description;
   // this directory is not neccessarily c:\windows
   LinkPath = GetWindowsDir()+"\\desktop";
   if (!DirectoryExists(LinkPath)) ForceDirectories(LinkPath);
   if (!DirectoryExists(LinkPath))
   {
      BingBox(this,
         s("Unable to locate the windows 'desktop' directory.\n")+
         s("I'll continue with the install, but without\n")+
         s("installing the desktop shortcut."),
         s("Problem"),
         MB_OK
      );
      return true;
   }
   AppName = ExtractFileName(AppExeName).UpperCase();
   AppName = AppName.SubString(1,AppName.Pos(".")-1);

   TargetPath = AppExeName;
   LinkPath += "\\Spend.LNK";
   IconPath = TargetPath;
   Description = "HashCash Spend";
   CreateLink(TargetPath,LinkPath,IconPath,Description,"-spend");

   LinkPath = GetWindowsDir()+"\\desktop";
   TargetPath = AppExeName;
   LinkPath += "\\Mint.LNK";
   IconPath = TargetPath;
   Description = "HashCash";
   CreateLink(TargetPath,LinkPath,IconPath,Description,"");

   MsgBoxTsk("Setup is complete!                              ","HashCash Setup",MB_OK);
   return true;
}
//---------------------------------------------------------------------------
/*bool __fastcall THashCashDlg::ExpiredToken(AnsiString token)
{
   AnsiString date=GetToken(1,token,":");
   int expired=ExpireDays->Text.ToIntDef(0);
   //  anon period needs work
   //int anonper=AnonDays->Text.ToIntDef(0);
   //
   int y=2000+date.SubString(1,2).ToIntDef(0);
   int m=date.SubString(3,2).ToIntDef(0);
   int d=date.SubString(5,2).ToIntDef(0);
   TDateTime toktime((unsigned short)y,(unsigned short)m,(unsigned short)d),nowtime=Date();
   if ((int)nowtime-(int)toktime>expired) return true;
   return false;
}*/
//---------------------------------------------------------------------------
int __fastcall THashCashDlg::TokenStatus(AnsiString token)
{
   char token_utime[MAX_UTC];
   char token_resource[MAX_RES];
   hashcash_parse(token.c_str(),token_utime,MAX_UTC,token_resource,MAX_RES);
   time_t token_time = from_utctimestr(token_utime);
   time_t nowtime = local_to_utctime(time(NULL));
   return hashcash_valid_for(token_time,ValidTimeInSeconds(),nowtime);
}
//---------------------------------------------------------------------------
/*bool __fastcall THashCashDlg::TokenValidInFuture(AnsiString token)
{
   return TokenStatus(token)==HASHCASH_VALID_IN_FUTURE;
}
//---------------------------------------------------------------------------
bool __fastcall THashCashDlg::ExpiredToken(AnsiString token)
{
   return TokenStatus(token)==HASHCASH_EXPIRED;
}*/
//---------------------------------------------------------------------------
AnsiString __fastcall THashCashDlg::ValidityPeriod()
{
   return ExpireTime->Text;
}
//---------------------------------------------------------------------------
AnsiString __fastcall THashCashDlg::AnonPeriod()
{
   return AnonTime->Text;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::ShowValidRandAvailValues()
{
   MintDisplay->Lines->Add("valid period: "+ValidityPeriod());
   if (AnonTimeInSeconds()) MintDisplay->Lines->Add("rand period: "+AnonPeriod());
   MintDisplay->Lines->Add("tokens available: "+ s(Pool->Items->Count));
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::CheckDeleteExpireds()
{
   int e=0,f=0;
   TStringList *ini = new TStringList();
   int i=0; while(i<Pool->Items->Count)
   {
      switch (TokenStatus(Pool->Items->Strings[i]))
      {
         case HASHCASH_EXPIRED: e++; Pool->Items->Delete(i); break;
         case HASHCASH_VALID_IN_FUTURE: f++;
         default: i++;
      }
   }
   if (f) Display(s(f)+" token"+(f>1?"s":"")+" valid in the future");
   if (e) Display("deleted "+s(e)+" expired token"+(e>1?"s":""));
   ShowValidRandAvailValues();
   delete ini;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::FormCloseQuery(TObject *Sender, bool &CanClose)
{
   if (CloseBtn->Caption=="Cancel")
   {
      PageControl->ActivePage = TabMint;
      if (BingBox(this,"Cancel minting operation?","Confirm",MB_YESNO)==IDNO)
      {
          CanClose = false;
          return;
      }
      Timer2->Enabled = false;
      // reset to org values whether attended or not
      QtyEdit->Text = holdtokens;
      BitsEdit->Text = holdbits;
      Resource->Text = holdresource;
      PriorityBox->ItemIndex = holdpriority;

      EnableInputs(true);
      ActiveControl = MintDisplay;
      ModalResult = mrNone;
      CanClose = false;
      attended = true;
      cancel = true;
      return;
   }
   Timer2->Enabled = false;
   if (modified) SaveTokenPool(Pool->Items);
   //{
   //   if (attended && ModalResult!=mrOk)
   //   {
   //      switch(MsgDlg(this,
   //         "Your token queue is modified. Save changes?",
   //         "Confirm",MB_YESNOCANCEL))
   //      {
   //         case IDCANCEL: CanClose = false; return;
   //         case IDNO: Snooze(200); goto SaveIni;
   //      }
   //   }
   //   SaveTokenPool(Pool->Items);
   //   Snooze(200);
   //}

   if (!attended) return;
   TIniFile *ini = new TIniFile(AppIni);
   ini->WriteString( "main","windowpos",s((int)centered)+","+s(Left)+","+s(Top));
   ini->WriteString( "main","tokens",holdqty.IsEmpty()?QtyEdit->Text:holdqty);
   ini->WriteInteger("main","bits",BitsEdit->Text.ToIntDef(17));
   ini->WriteString( "main","resource",Resource->Text);
   ini->WriteString( "main","priority",PriorityBox->Text.LowerCase());
   ini->WriteString( "main","expires",ValidityPeriod());
   ini->WriteString( "main","randperiod",AnonPeriod());
   ini->WriteString( "main","timewidth",TimeWide->Text);
   ini->WriteInteger("main","windowtextcolor",WColorBox->ItemIndex);
   ini->WriteInteger("main","inputtextcolor",IColorBox->ItemIndex);
   delete ini;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::Timer1Timer(TObject *Sender)
{
   static c=0;
   switch(++c)
   {
      case 1:
         Caption = "Unattended Mode...";
         PageControl->ActivePage = TabMint;
         OkBtn->Click();
         break;
      case 3:
         SetRedraw(this,1);
         if (clMinimized) Caption = "Minimizing...";
         Snooze(1);
         break;
      case 6:
         Timer1->Enabled = false;
         if (clMinimized) Application->Minimize();
         Caption = cap;
         break;
   }
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::TimeBoxClick(TObject *Sender)
{
   bool calc=TimeBox->Checked;
   QtyEdit->Enabled = QtyUpDn->Enabled = Label1->Enabled = !calc;
   if (!calc)
   {
      QtyEdit->Text = holdqty; holdqty = "";
      OkBtn->Caption = "Mint";
   }
   else
   {
      holdqty = QtyEdit->Text; QtyEdit->Text = "0";
      OkBtn->Caption = "Time";
   }
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::DisplayLast(AnsiString line)
{
   int c =MintDisplay->Lines->Count;
   if (c) MintDisplay->Lines->Strings[c-1] = line;
   else Display(line);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::Display(AnsiString line)
{
   ActiveControl = MintDisplay;
   MintDisplay->Lines->Add(line);
   ActiveControl = Panel1;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::PoolClick(TObject *Sender)
{
   //DeleteBtn->Enabled = Pool->SelCount;
   OkBtn->Enabled = Pool->SelCount==1;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::PageControlChange(TObject *Sender)
{
   if (PageControl->ActivePage==TabPool)
   {
      if (!Pool->SelCount) ActiveControl = Panel1;
      Label12->Caption = s(Pool->Items->Count)+" tokens";
      OkBtn->Enabled = Pool->SelCount==1;
      OkBtn->Caption = "Spend";
      return;
   }
   if (PageControl->ActivePage==TabMint)
   {
      ActiveControl = QtyEdit;
      OkBtn->Caption = "Mint";
      OkBtn->Enabled = true;
      return;
   }
   if (PageControl->ActivePage==TabOpts)
   {
      ActiveControl = ExpireTime;
      OkBtn->Caption = "Mint";
      OkBtn->Enabled = true;
      return;
   }
}
//---------------------------------------------------------------------------
AnsiString __fastcall Seconds2HMS(double Seconds)
{
   double rem,scs,mns,hrs,seconds,Minutes,minutes,Hours,hours;
   rem = modf(Seconds,&seconds);

   Minutes = seconds / 60;
   modf(Minutes,&minutes);
   scs = seconds - (minutes * 60);
   if (rem>=.5) scs++;

   Hours = minutes / 60;
   modf(Hours,&hours);
   mns = minutes - (hours * 60);

   hrs = hours;

   AnsiString h=(int)hrs,m=(int)mns,s=(int)scs;
   if (h.Length()<2) h = "0"+h;
   if (m.Length()<2) m = "0"+m;
   if (s.Length()<2) s = "0"+s;
   return (h+":"+m+":"+s);
}
//---------------------------------------------------------------------------
TThreadPriority __fastcall THashCashDlg::Priority()
{
   switch(PriorityBox->ItemIndex)
   {
      case 0: return tpTimeCritical;
      case 1: return tpHighest;
      case 2: return tpHigher;
      case 3: return tpNormal;
      case 4: return tpLower;
      case 5: return tpLowest;
      case 6: return tpIdle;
   }
   return tpNormal;
}
//---------------------------------------------------------------------------
BOOL __fastcall THashCashDlg::GetIniSection(AnsiString name, TStrings *section,
   TStrings *ini)
{
   AnsiString line;
   section->Clear();
   bool insection=false;
   for (int i=0; i<ini->Count; i++)
   {
      line = ini->Strings[i].Trim();
      if (line.LowerCase()=="["+name.LowerCase()+"]")
      {
         insection = true;
         continue;
      }
      if (insection)
      {
         if (line.IsEmpty()) break;
         section->Add(line);
      }
   }
   return section->Count>0;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::LoadTokenPool(TStrings *tokens)
{
   if (FileExists(AppQue)) tokens->LoadFromFile(AppQue);
   else tokens->Clear();
   modified = false;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::SaveTokenPool(TStrings *tokens)
{
   tokens->SaveToFile(AppQue);
   modified = false;
}
//---------------------------------------------------------------------------
/*void __fastcall THashCashDlg::SaveTokenPool(TStrings *tokens)
{
   TStringList *ini = new TStringList();
   TStringList *main = new TStringList();
   if (FileExists(AppIni)) ini->LoadFromFile(AppIni);
   GetIniSection("main",main,ini);
   ini->Clear();
   ini->Add("[main]");
   ini->AddStrings(main);
   ini->Add("");
   ini->Add("[queue]");
   ini->AddStrings(tokens);
   ini->SaveToFile(AppIni);
   delete main;
   delete ini;
}*/
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::CloseBtnClick(TObject *Sender)
{
   ModalResult = mrOk;
   Close();
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::EnableInputs(bool Enable)
{
   for (int i=0; i<TabMint->ControlCount; i++)
   {
      TControl *control = TabMint->Controls[i];
      if (control!=MintDisplay && control!=Label8) control->Enabled = Enable;
   }
   CloseBtn->Caption = Enable ? "Close":"Cancel";
   MintPanel->Caption = !Enable ? "Minting HashCash Tokens":"";
   if (TimeBox->Checked) MintPanel->Caption = "";
   MintPanel->Visible = !Enable;
   TabOpts->TabVisible = Enable;
   TabPool->TabVisible = Enable;
   HelpBtn->Enabled = Enable;
   OpenBtn->Enabled = Enable;
   OkBtn->Enabled = Enable;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TwoDigits(int value)
{
   AnsiString result = s(abs(value));
   if (result.Length()==1) result = "0"+result;
   return result;
}
//---------------------------------------------------------------------------
AnsiString __fastcall Seconds2DHMS(long seconds)
{
   double dys,hrs,mns,scs,secs=abs(seconds);
   hrs = fmod(secs,86400); dys = (secs-hrs)/86400;
   mns = fmod(hrs,3600); hrs = (hrs-mns)/3600;
   scs = fmod(mns,60); mns = (mns-scs)/60;
   return s(seconds<0?"-":"+")+TwoDigits(dys)+":"+TwoDigits(hrs)+":"+
      TwoDigits(mns)+":"+TwoDigits(scs);
}
//---------------------------------------------------------------------------
AnsiString __fastcall RandomOffset(long seconds)
{
   return "random offset: "+Seconds2DHMS(seconds);
}
//---------------------------------------------------------------------------
int __fastcall THashCashDlg::ValidField(int pos)
{
   if (pos>=0 && pos<4) return GetToken(pos,ExpireTime->Text,":").ToIntDef(0);
   return 0;
}
//---------------------------------------------------------------------------
int __fastcall THashCashDlg::ValidTimeInSeconds()
{
   AnsiString valid=ExpireTime->Text.LowerCase();
   if (!valid.IsEmpty() && valid.Pos("s")==valid.Length())
      return valid.SubString(1,valid.Length()-1).ToIntDef(0);
   return (86400*ValidField(0))+(3600*ValidField(1))+(60*ValidField(2))+ValidField(3);
}
//---------------------------------------------------------------------------
int __fastcall THashCashDlg::AnonField(int pos)
{
   if (pos>=0 && pos<4) return abs(GetToken(pos,AnonTime->Text,":").ToIntDef(0));
   return 0;
}
//---------------------------------------------------------------------------
int __fastcall THashCashDlg::AnonTimeInSeconds()
{
   AnsiString anon; int result;
   bool isminus=AnonTime->Text.Pos("-")==1;
   if (isminus) anon = AnonTime->Text.SubString(2,32).LowerCase();
   else anon = AnonTime->Text.LowerCase();
   if (!anon.IsEmpty() && anon.Pos("s")==anon.Length())
   {
      result = anon.SubString(1,anon.Length()-1).ToIntDef(0);
      if (isminus) result = -result;
      return result;
   }
   result = (86400*AnonField(0))+(3600*AnonField(1))+(60*AnonField(2))+AnonField(3);
   if (isminus) result = -result;
   return result;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::OkBtnClick(TObject *Sender)
{
   if (attended && OkBtn->Caption=="Spend")
   {
       int i=Pool->ItemIndex; if (i<0) return;
       AnsiString token=Pool->Items->Strings[i];
//     if (TokenStatus(token)==HASHCASH_EXPIRED) sm("EXPIRED");
       if (TokenStatus(token)==HASHCASH_VALID_IN_FUTURE)
       {
         AnsiString validdate = token.SubString(1,2)+"-"+token.SubString(3,2)+"-"+token.SubString(5,2);
         BingBox(this,
            s("The token you have selected is not yet valid. It's creation date is: "+validdate+"\n\n")+
            s("Because the queue is sorted in date order, all other tokens you may have are\n")+
            s("also valid in the future. You will need to mint some before you can\n")+
            s("continue."),
            s("Token Not Yet Valid"),MB_OK);
         return;
      }
      if (Pool->Items->Count && Pool->ItemIndex>=0)
      {
         Clipboard()->AsText = Xhashcash+Pool->Items->Strings[Pool->ItemIndex];
         Pool->Items->Delete(Pool->ItemIndex);
         if (Pool->Items->Count)
         {
            //Pool->Selected[0] = true;
            //Pool->ItemIndex = 0;
         }
         else OkBtn->Enabled = false;
      }
      return;
   }

   if (!ValidInput()) return;

   Snooze(100); PageControl->ActivePage = TabMint; Snooze(100);
   //sm(s((int)_timezone)+"   "+s((int)time(NULL)));
   EnableInputs(false);
   int bits = BitsEdit->Text.ToIntDef(17);
   MintDisplay->Clear(); Snooze(1);
   Display("working...");
   AnsiString priority=PriorityBox->Text,resource=Resource->Text;
   int retvalue,tokens=QtyEdit->Text.ToIntDef(1);

   // save whether attended or not
   holdtokens = QtyEdit->Text;
   holdbits = BitsEdit->Text;
   holdresource = Resource->Text;
   holdpriority = PriorityBox->ItemIndex;

   if (!attended)
   {
      // change values that may need changing from command line
      if (clTokens>0) tokens = clTokens;
      if (clResource!="") resource = clResource;
      if (clPriority!="") priority = clPriority;
      if (clBits>0) bits = clBits;
      // list all values used first
      DisplayLast("mode=unattended");
      Display("tokens="+s(tokens));
      Display("bits="+s(bits));
      Display("resource="+resource);
      Display("priority="+priority);
      Display("");
      // assign curr values to display edits
      QtyEdit->Text = tokens;
      BitsEdit->Text = bits;
      Resource->Text = resource;
      PriorityBox->ItemIndex = PriorityBox->Items->IndexOf(priority);
      if (PriorityBox->ItemIndex<0) PriorityBox->Items->IndexOf("normal");
      Snooze(10);
   }
   DisplayLast("started "+FormatDateTime("mm/dd hh:nn:ss", Now()));
   Display("");

   AnsiString esttime;
   DoEstimateTimeRequired(bits,Priority(),&esttime);
   if (TimeBox->Checked)
   {
      //CloseBtn->Caption = "Close";
      //OkBtn->Enabled = true;
      DisplayLast(esttime);
      EnableInputs(true);
      return;
   }
   minted = 0;
   ActiveControl = MintDisplay;
   long anon_period = AnonTimeInSeconds();
   time_t validity_period = ValidTimeInSeconds();
   MintPanel->Caption = "Minting HashCash Tokens";
   int timewidth = TimeWide->Text.ToIntDef(6);
   for (int i=0; i<tokens; i++)
   {
      if (cancel) break;
      hs = ms = ss = 0;
      long anon_random=0;
      double tries_taken=0;
      TimePanel->Caption = "00:00:00";
      TimePanel->Visible = true;
      Timer2->Enabled = true;
      double ticks = GetTickCount();
      if (MintDisplay->Lines->Count==1) MintDisplay->Lines->Delete(0);
      MintPanel->Caption = "Minting HashCash Token: "+s(minted+1)+" of "+s(tokens);
      Display(esttime);
      Application->Title = "HCW  "+s(i+1)+" of "+s(tokens);
      Snooze(200); DisplayLast("working..."); Snooze(300);
      Timer2->Enabled = true;

      AnsiString token;
      switch(retvalue=DoMintToken(timewidth,bits,resource,validity_period,anon_period,
         Priority(),token,&anon_random,&tries_taken,&cancel))
      {
         case HASHCASH_OK:
            ticks = GetTickCount()-ticks;
            DisplayLast("actual time: "+s((int)(ticks/1000))+" seconds ("+
               Seconds2HMS((int)(ticks/1000))+")");
            Display("token: "+token);
            if (anon_random!=0) Display(RandomOffset(anon_random));
            Display("tries taken: "+s(tries_taken));
            if (i<tokens-1) Display("");

            CreateEmptyFile(AppQueLck);
            LoadTokenPool(Pool->Items);
            Pool->Items->Add(token);
            SaveTokenPool(Pool->Items);
            DeleteFile(AppQueLck.c_str());
            modified = true;
            minted++;
            break;
         case HASHCASH_USERCANCEL:
            ticks = GetTickCount() - ticks;
            DisplayLast("actual time: "+s((int)(ticks/1000))+" seconds ("+
               Seconds2HMS((int)(ticks/1000))+")");
            Display("user cancelled");
            break;
         case HASHCASH_INSUFFICIENT_BITS:
            DisplayLast("error: insufficient bits for minting");
            Display("increase bit count");
            break;
         case HASHCASH_TOO_MANY_TRIES:
            Display("error: token requires too many calculations");
            Display("decrease the bit count");
            break;
         case HASHCASH_EXPIRED_ON_CREATION:
            Display("error: token is already expired!");
            break;
         case HASHCASH_FAIL:
         case HASHCASH_EXPIRED:
         case HASHCASH_INVALID_TOK_LEN:
         case HASHCASH_INVALID_TIME:
         case HASHCASH_RNG_FAILED:
         case HASHCASH_INVALID_VALIDITY_PERIOD:
         case HASHCASH_NULL_ARG:
         case HASHCASH_INVALID_TIME_WIDTH:
         case HASHCASH_VALID_IN_FUTURE:
         case HASHCASH_INVALID:
         case HASHCASH_WRONG_RESOURCE:
            DisplayLast("internal error, can't mint token");
            break;
      }
      Label8->Caption = s(minted)+(minted==1?" new token":" new tokens");
      if (retvalue!=HASHCASH_OK)
         break;

      TimePanel->Caption = "";
      Snooze(500);
   }
   Display("\nfinished "+FormatDateTime("mm/dd hh:nn:ss", Now()));
   MintDisplay->Lines->SaveToFile(ChangeFileExt(AppExeName,".log"));
   if (minted==tokens && !cancel) Application->Title = "HCW done";
   else Application->Title = "HCW cancelled";

   TimePanel->Visible = false;
   PageControl->ActivePage = TabMint;
   ActiveControl = Panel1;
   cancel = false;
   EnableInputs(true);
   if (attended) MessageBeep(MB_OK);
   else
   {
      Display("\nclosing...");
      Snooze(500);
      Close();
   }
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::RememberBtnClick(TObject *Sender)
{
   centered = false;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::CenterBtnClick(TObject *Sender)
{
   centered = true;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::WColorBoxDrawItem(TWinControl *Control, int Index,
	TRect &Rect, TOwnerDrawState State)
{
   TCanvas *canvas = WColorBox->Canvas;
   canvas->Brush->Color = GetColor16(Index);
   canvas->Rectangle(Rect.Left+2,Rect.Top+1,Rect.Right-2,Rect.Bottom-1);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::IColorBoxDrawItem(TWinControl *Control, int Index,
	TRect &Rect, TOwnerDrawState State)
{
   TCanvas *canvas = IColorBox->Canvas;
   canvas->Brush->Color = GetColor16(Index);
   canvas->Rectangle(Rect.Left+2,Rect.Top+1,Rect.Right-2,Rect.Bottom-1);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::FormPaint(TObject *Sender)
{
   TRect Rct = Rect(4,0,MintDisplay->ClientWidth-2,ClientHeight);
   MintDisplay->Perform(EM_SETRECT,0,long(&Rct));
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::FormResize(TObject *Sender)
{
   TRect Rct = Rect(4,0,MintDisplay->ClientWidth-2,ClientHeight);
   MintDisplay->Perform(EM_SETRECT,0,long(&Rct));
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::WColorBoxChange(TObject *Sender)
{
   MintDisplay->Font->Color = (TColor)GetColor16(WColorBox->ItemIndex);
   Pool->Font->Color = (TColor)GetColor16(WColorBox->ItemIndex);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::IColorBoxChange(TObject *Sender)
{
   TColor color = GetColor16(IColorBox->ItemIndex);
   PriorityBox->Font->Color = color;
   BitsEdit->Font->Color = color;
   Resource->Font->Color = color;
   QtyEdit->Font->Color = color;
   ExpireTime->Font->Color = color;
   AnonTime->Font->Color = color;
   TimeWide->Font->Color = color;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::AboutBtnClick(TObject *Sender)
{
   AboutBox(this);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::HelpBtnClick(TObject *Sender)
{
   int page = 0;
   if (PageControl->ActivePage==TabMint) page = 1; else
   if (PageControl->ActivePage==TabPool) page = 2; else
   if (PageControl->ActivePage==TabOpts) page = 3;
   OpenHelp(this,page,centered);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::Timer2Timer(TObject *Sender)
{
   ss++;
   if (ss==60)
   {
      ss = 0;
      ms++;
   }
   if (ms==60)
   {
      ms = 0;
      hs++;
   }
   AnsiString hstr,mstr,sstr;
   hstr = s(hs); if (hstr.Length()==1) hstr = "0"+hstr;
   mstr = s(ms); if (mstr.Length()==1) mstr = "0"+mstr;
   sstr = s(ss); if (sstr.Length()==1) sstr = "0"+sstr;
   TimePanel->Caption = hstr+":"+mstr+":"+sstr;
   Snooze(10);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::PoolDrawItem(TWinControl *Control, int Index,
	TRect &Rect, TOwnerDrawState State)
{
   TCanvas *canvas = Pool->Canvas; canvas->FillRect(Rect);
   AnsiString token = Pool->Items->Strings[Index];
 //if (TokenStatus(token)==HASHCASH_VALID_IN_FUTURE) canvas->Font->Color=clHighLighText;
   canvas->TextRect(Rect,Rect.Left+4,Rect.Top,token);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::hcwlog1Click(TObject *Sender)
{
   ViewMintLog(this,MintDisplay->Font->Color);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::hcwini1Click(TObject *Sender)
{
   EditAppIni(this,MintDisplay->Font->Color);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::TimeChange(TObject *Sender)
{
   MintDisplay->Clear();
   ShowValidRandAvailValues();
   Label8->Caption = "";
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::SaveBtnClick(TObject *Sender)
{
   AnsiString cap = Caption;
   Caption = "Saving..."; Snooze(400);
   SaveTokenPool(Pool->Items);
   SaveBtn->Enabled = false;
   modified = false;
}
//---------------------------------------------------------------------------
HANDLE __fastcall Exec95QSnews(TForm *Owner, const AnsiString& strCmdLine,
   const WORD ShowCmd)
{
   STARTUPINFO siStartInfo; 	  // Set up members of STARTUPINFO structure
   PROCESS_INFORMATION piProcInfo;
   siStartInfo.cb = sizeof(STARTUPINFO);
   siStartInfo.lpReserved = NULL;
   siStartInfo.lpDesktop = NULL;
   siStartInfo.lpTitle = NULL;
   siStartInfo.wShowWindow = ShowCmd;
   siStartInfo.lpReserved2 = NULL;
   siStartInfo.cbReserved2 = 0;
   siStartInfo.dwFlags = STARTF_USESHOWWINDOW;
   if (!CreateProcess(NULL,   	  // Create the child process
         strCmdLine.c_str(),
         NULL, 	                  // process security attributes
         NULL, 					  // primary thread security attributes
         0, 					  // handles are inherited
         0, 					  // creation flags
         NULL, 					  // use parent's environment
         NULL, 					  // use parent's current directory
        &siStartInfo, 			  // STARTUPINFO pointer
        &piProcInfo))
   {
      BingBox(Owner,
         s("I encountered this error when starting the QSnews module: \n")+
         s(GetLastErrorStr()+"\n")+
         s("If the problem persists please send me a bug report and\n")+
         s("and include the error message.\n"),
         s("System Error"),MB_OK);
   }
   return piProcInfo.hProcess;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::PopupMenu3Popup(TObject *Sender)
{
   bool c = Pool->SelCount;
   PopImportItem->Enabled = Clipboard()->HasFormat(CF_TEXT);
   PopExportItem->Enabled = c>0;
   PopDeleteItem->Enabled = c>0;
}
//---------------------------------------------------------------------------
bool __fastcall THashCashDlg::ValidHashCash(AnsiString &token, AnsiString &errormsg)
{
   bool ok; int c=0;
   for (int i=0; i<token.Length(); i++) if (token[i]==':') c++;
   if (c<2)
   {
      errormsg = "It does not contain 3 fields separated by 2 colons.";
      return false;
   }
   ok=true; for (int i=1; i<token.Length(); i++)
   {
      if (token[i]==':') break;
      if (s("0123456789").Pos(token[i])==0) ok = false;
      if (!ok) break;
   }
   if (!ok)
   {
      errormsg = "The date field, contains non-digit character(s).";
      return false;
   }
   ok=true; for (int i=token.Length(); i>1; i--)
   {
      if (token[i]==':') break;
      if (s("0123456789ABCDEFabcdef").Pos(token[i])==0) ok = false;
      if (!ok) break;
   }
   if (!ok)
   {
      errormsg = "The last field, contains non-hexdigit character(s).";
      return false;
   }
   if (TokenStatus(token)==HASHCASH_EXPIRED)
   {
      errormsg = "It is expired!";
      return false;
   }
   return true;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::PopExportItemClick(TObject *Sender)
{
   AnsiString filename;
   if (!Pool->Items->Count) return;
   if (!ExportHashCashTokens(this,filename)) return;

   bool modified=false;
   Pool->Items->BeginUpdate();
   TStringList *exports = new TStringList();
   int i=0; while (i<Pool->Items->Count)
   {
      if (Pool->Selected[i])
      {
         exports->Add(Pool->Items->Strings[i]);
         Pool->Items->Delete(i);
         modified = true;
      }
      else i++;
   }
   if (modified)
   {
      Label12->Caption = s(Pool->Items->Count)+" items";
      if (filename.IsEmpty()) Clipboard()->AsText = exports->Text;
      else exports->SaveToFile(filename);
      SaveTokenPool(Pool->Items);
   }
   Pool->Items->EndUpdate();
   delete exports;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::PopImportItemClick(TObject *Sender)
{
   TStringList *tokens = new TStringList();
   bool doimport=ImportHashCashTokens(this,tokens);
   if (!doimport) return;
   if (!tokens->Count)
   {
      MsgDlg(this,"No tokens to import!         ","Import Tokens",MB_OK);
      delete tokens;
      return;
   }

   Pool->Items->BeginUpdate();
   for (int i=0; i<Pool->Items->Count; i++)
      Pool->Selected[i] = false;
   //Pool->ItemIndex = 0;

   bool modified = false;
   AnsiString errmsg,token;
   for (int i=0; i<tokens->Count; i++)
   {
      token = tokens->Strings[i];
      if (!ValidHashCash(token,errmsg))
      {
         switch(MsgDlg(this,
            s("On examining this token I found it invalid.\n"+errmsg)+
            s("\n\n"+token+"\n\nDiscard this token and continue?"),
            s("Error Reading Tokens"),
            MB_OKCANCEL))
         {
            case IDOK: continue;
            case IDCANCEL:
               delete tokens;
               return;
         }
      }
      int idx=Pool->Items->Add(token);
      if (idx>=0) Pool->Selected[idx] = true;
      modified = true;
   }
   if (modified)
   {
      Label12->Caption = s(Pool->Items->Count)+" items";
      SaveTokenPool(Pool->Items);
   }
   Pool->Items->EndUpdate();
   delete tokens;
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::TokensItemClick(TObject *Sender)
{
   TPoint screen,client={12+TokensBtn->Left,32+TokensBtn->Top+TokensBtn->Height};
   screen = ClientToScreen(client);
   PopupMenu3->Popup(screen.x,screen.y);
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::PoolDblClick(TObject *Sender)
{
   OkBtn->Click();
}
//---------------------------------------------------------------------------
void __fastcall THashCashDlg::PopDeleteItemClick(TObject *Sender)
{
    if (ConfirmBox(this,"Delete the selected tokens?",MB_OKCANCEL)==IDOK)
    {
       bool modified=false;
       Pool->Items->BeginUpdate();
       int i=0; while(i<Pool->Items->Count)
       {
          if (Pool->Selected[i]) { Pool->Items->Delete(i); modified = true; }
          else i++;
       }
       //if (Pool->Items->Count)
       //{
       //   Pool->Selected[0] = true;
       //   Pool->ItemIndex = 0;
       //}
       //Label7->Caption = "HashCash Tokens";//, "+s(Pool->Items->Count);
       if (modified)
       {
          Label12->Caption = s(Pool->Items->Count)+" items";
          SaveTokenPool(Pool->Items);
       }
       Pool->Items->EndUpdate();
    }
}
//---------------------------------------------------------------------------
